<?php
//C:\xampp\htdocs\sanskrit\transcoder\transcoder.php
// the $transcoder_dir global is used to determine
//$transcoder_dir="C:/xampp/htdocs/sanskrit/transcoder";
//$transcoder_dir="c:/server2go/htdocs/sanskrit/transcoder";
$transcoder_dir="transcoder"; // default

$transcoder_fsmarr = array();
function transcoder_fsm($from,$to) {
 global $transcoder_dir,$transcoder_fsmarr;
 $fromto = $from . "_" . $to;
 if ($transcoder_fsmarr[$fromto]) {
  return;
 }
 $filein = $transcoder_dir . "/" . $fromto . ".xml";
 $xml = simplexml_load_file($filein);
 $entries = $xml->xpath('*');
 $n = count($entries);
 $start = (string) $xml['start'];
 $fsm=array();
 $fsm['start']=$start;
 $fsmentries = array();
 foreach($entries as $e) {
  $x = $e->xpath('in');
  $inval=(string)$x[0];
  $conlook=FALSE;
  if (preg_match('/^(.)\/\^/',$inval)) {
   // In transcoding from slp1 to devanagari, it is necessary to do a
   // 'look-ahead' when deciding how to code a consonant.  If the 
   // consonant is not followed by a vowel, then a vigraha has to be emitted.
   // The input codes $inval in such cases as:
   // k/^([^aAiIuUfFxXeEoO^/\\])
   // Which is to be intepreted as: starting at the next character,
   //    check if the input string does NOT match the regular expression
   //    [^aAiIuUfFxXeEoO^/\\].
   //    Note that the last 3 elements '^', '/', and '\' are present only
   //    because of accents. 
   if (! ($fromto == 'slp1_deva')) {continue;}
   $inval = $inval[0];
   $conlook=TRUE;
  }
  $x = $e -> xpath('s');
  $sval = (string) $x[0];
  $x = $e->xpath('out');
  $outval=(string) $x[0];
  $x = $e->xpath('next');
  $nextval=(string) $x[0];

  $startStates = preg_split('/,/',$sval,-1, PREG_SPLIT_NO_EMPTY);
  if ($nextval == '') {
   $nextval = $startStates[0];
  }
  $nextState = $nextval;
   $utf="";
   $vals=array();
   $u1 = $outval;
    if (substr($u1,0,2) == '\u') {
     $us=preg_split('/\\\\u/',$u1,-1, PREG_SPLIT_NO_EMPTY);
     $v='';
     foreach($us as $u) {
      $v=unichr(hexdec($u));
      $utf .= $v;
     }
     $vals[]=$utf;
    }else {
     $vals[]=$u1;
    }
   $newoutval = implode('',$vals);
   $fsmentry=array();
   $fsmentry['starts']=$startStates;
   $fsmentry['in']=$inval;
   if ($conlook) {
    $fsmentry['regex']=TRUE;
    //otherwise, regex is undefined
   }
   $fsmentry['out']=$newoutval;
   $fsmentry['next']=$nextState;
   $fsmentries[]=$fsmentry;
 }
 $fsm['fsm']=$fsmentries;
// make associative array $states, whose keys are characters,
// and whose value at a key is an array of subscripts into $fsmentries.
//  $i is a subscript for a key provided that the $fsmentries[$i]['in'] = 
//  first character of $key
 $states=array();
 foreach($fsmentries as $i => $fsmentry) {
  $in = $fsmentry['in'];
  $c = $in[0];
  $state=$states[$c];
  if ($state) {
    $state[]=$i;
    $states[$c]=$state;
  }else {
   $state = array();
   $state[]=$i;
   $states[$c]=$state;
  }
 }
 $fsm['states']=$states;
 $transcoder_fsmarr[$fromto]=$fsm;
} 
function unichr($dec) {
  if ($dec < 128) {
    $utf = chr($dec);
  } else if ($dec < 2048) {
    $utf = chr(192 + (($dec - ($dec % 64)) / 64));
    $utf .= chr(128 + ($dec % 64));
  } else {
    $utf = chr(224 + (($dec - ($dec % 4096)) / 4096));
    $utf .= chr(128 + ((($dec % 4096) - ($dec % 64)) / 64));
    $utf .= chr(128 + ($dec % 64));
  }
  return $utf;
}
function transcoder_processString($line,$from,$to) {
 global $transcoder_dir,$transcoder_fsmarr;
 if ($from == $to) {return $line;}
 $fromto = $from . "_" . $to;
 $fsm = $transcoder_fsmarr[$fromto];
 if (!$fsm) {
  transcoder_fsm($from,$to);
  $fsm = $transcoder_fsmarr[$fromto];
  if (!$fsm) {
   return $line;
  }
 }
// echo "found fsm for $fromto\n";
 $currentState=$fsm['start'];
 $fsmentries = $fsm['fsm'];
 $states = $fsm['states'];
 $n=0; // current character position in $line
 $result=''; // returned value
 $m=strlen($line);
 while ($n < $m) {
  $c = $line[$n];
  $isubs = $states[$c];
  if (! $isubs) {
   $result .= $c;
//   echo "chk: '$c' no states, n=$n<br>\n";
   $n++;
   continue;
  }
  $best="";
  $nbest=0;
  $bestFE = NULL;
  foreach($isubs as $isub) {
   $fsmentry=$fsmentries[$isub];
   $startStates=$fsmentry['starts'];
   $k=-1;
   $nstartStates=count($startStates);
   for($j=0;$j<$nstartStates;$j++) {
    if ($startStates[$j] == $currentState) {
     $k=$j;
     $j=$nstartStates;
    }
   }
   if ($k == -1) {continue;}
   $match = transcoder_processString_match($line,$n,$m,$fsmentry);
   $nmatch=strlen($match);
//   echo "chk2: n=$n, c='$c', nmatch=$nmatch<br>\n";
   if ($nmatch > $nbest) {
    $best = $match;
    $nbest=$nmatch;
    $bestFE=$fsmentry;
   }
  }
  if ($bestFE) {
   $result .= $bestFE['out'];
   $n += $nbest;
   $currentState=$bestFE['next'];
  }else {
//   echo "chk1: n=$n, c='$c' state=$currentState: no best<br>\n";
   $result .= $c;
   $n++;
  }
 }
 return $result;
}
function transcoder_processString_match($line,$n,$m,$fsmentry) {
  $match=""; // value returned
  $edge = $fsmentry['in'];
  $nedge=strlen($edge);
  $j=$n;
  $k=0;
  $b=TRUE;
  while ( ($j < $m) && ($k < $nedge) && $b) {
   if($line[$j] == $edge[$k]) {
    $j++;
    $k++;
   }else {
    $b=FALSE;
   }
  }
  if (!$b) { return $match;}
  if ($k != $nedge)  { return $match;}
  $match=$edge;
  if (!$fsmentry['regex']) {
   return $match;
  }
  //  additional logic when $fsmentry['regex'] is TRUE
  //  see discussion of 'regex' in transcoder_fsm
  //  This logic only works with slp1_deva xml file.
  //  Also, it ignores the use of '/^\' as vowel accents.
  if ($n+1 == $m) {return $match;} 
  $d = $line[$n+1];
  if (preg_match('/[^aAiIuUfFxXeEoO^\/\\\\]/',$d)) {return $match;}
  return "";
}
function transcoder_processElements($line,$from,$to,$tagname){
 // Assume parts of $line to be converted are marked in an xml way.
 // For example, if $tagname = <SA>:
 //  and $line = 'The word <SA>rAma</SA> refers to a person',
 // returned would be 'The word XXX refers to a person',
 // where XXX is the transformation of the the string 'rAma' acc. to $from,$to

 $ans = preg_replace("/<$tagname>(.*?)<\/$tagname>/e",
         "transcoder_processString('\\1','$from','$to')",$line);
 return $ans;
}
function transcoder_standardize_filter($filter) {
 // standardizes a 'filter' into a transcoder standard.
 $standard_hash=array(
  "SktDevaUnicode"=>"deva",
  "SktRomanUnicode"=>"roman",
  "SLP2SLP"=>"slp1",
  "SLP2HK" =>"hk",
  "HK2SLP" =>"hk",
  "SLP2ITRANS" => "itrans",
  "ITRANS2SLP" => "itrans",
  "HK" =>"hk",
  "ITRANS" => "itrans",
  "HKT" => "hkt",
  "HKT2HKT" => "hkt",
  "TAMIL" => "tamil"
 );
$result=$standard_hash[$filter];
if (!$result) {$result="slp1";}
return $result;
}
function transcoder_set_dir($dir) {
 // may return FALSE if string $dir is improper in some way
 global $transcoder_dir;
 $transcoder_dir=realpath($dir);
 if (! file_exists($transcoder_dir)) {
  $transcoder_dir=FALSE;
 }
 return $transcoder_dir;
}
function transcoder_get_dir() {
 global $transcoder_dir;
 return $transcoder_dir;
}
?>